/****************************************Copyright (c)****************************************************
**                                      
**                                
**
**--------------File Info---------------------------------------------------------------------------------
** File name:               SPI_MSD_Driver.c
** Descriptions:            The SPI SD Card application function
**
**--------------------------------------------------------------------------------------------------------
** Modified by:             
** Modified date:           
** Version:                 
** Descriptions:            
**
*********************************************************************************************************/

/* Includes ------------------------------------------------------------------*/
#include <stdio.h>
#include "CH58x_common.h"
#include "SPI_MSD_Driver.h"
#include "ff.h"

/* Private define ------------------------------------------------------------*/
#define PRINT_INFO  1

/********************************* 引脚定义 ************************************
*    PA12  <===========>  CS
*    PA13  <===========>  SCK
*    PA14  <===========>  DI/MOSI
*    PA15  <===========>  DO/MISO
*    PA4    TF/SD Card Check
*******************************************************************************/
#define  SD_CS_PIN_LOW()        GPIOA_ResetBits( GPIO_Pin_12 )
#define  SD_CS_PIN_HIGH()       GPIOA_SetBits( GPIO_Pin_12 )
#define  SD_INSERT_PIN_IN()     GPIOA_ReadPortPin(GPIO_Pin_4)   /* SD卡插入引脚检测 */

/* Physical level marcos */
#define _card_enable()          GPIOA_ResetBits( GPIO_Pin_12 )
#define _card_disable()         GPIOA_SetBits( GPIO_Pin_12 )
#define _card_power_on()
#define _card_power_off()
#define _card_insert()          GPIOB_ReadPortPin(GPIO_Pin_21)   /* SD卡插入引脚检测 */


int _send_command(uint8_t cmd, uint32_t arg, uint8_t crc);
int _send_command_hold(uint8_t cmd, uint32_t arg, uint8_t crc);
int _read_buffer(uint8_t *buff, uint16_t len, uint8_t release);

/* Private variables ---------------------------------------------------------*/
MSD_CARDINFO cardInfo;

FATFS       fs;         /* Work area (file system object) for logical drive */
FIL         fsrc;         /* file objects */
FRESULT     res;
UINT        br;
char        path[128]="0:";
#ifdef DEBUG_INFO
UINT8       buf[1024];
char        textFileBuffer[] = "Thank you for using Development Board ^_^ \r\n";
#endif

/*******************************************************************************
* Function Name  : _spi_read_write
* Description    : None
* Input          : - data:
* Output         : None
* Return         : None
* Attention         : None
*******************************************************************************/
#define _spi_read_write(d)      SPI0_MasterSendRecvByte(d)

//__inline int _spi_read_write(uint8_t data)
//{
//  /* Return the byte read from the SPI2 bus */
//  return 0;
//}

/*******************************************************************************
* Function Name  : MSD_SPI_Configuration
* Description    : SD Card SPI Configuration
* Input          : None
* Output         : None
* Return         : None
* Attention         : None
*******************************************************************************/
void MSD_SPI_Configuration(void)
{
    GPIOA_SetBits( GPIO_Pin_12|GPIO_Pin_13|GPIO_Pin_14 );
    GPIOA_ModeCfg( GPIO_Pin_12|GPIO_Pin_13|GPIO_Pin_14, GPIO_ModeOut_PP_5mA );    // MOSI/SCK/CS
    GPIOA_SetBits( GPIO_Pin_15|GPIO_Pin_4 );
    GPIOA_ModeCfg( GPIO_Pin_15|GPIO_Pin_4, GPIO_ModeIN_PU );    // MISO

    SPI0_MasterDefInit();
    R8_SPI0_CTRL_MOD = RB_SPI_MOSI_OE | RB_SPI_SCK_OE | RB_SPI_MISO_OE ;
    SPI0_DataMode( Mode3_HighBitINFront );
    SPI0_CLKCfg( 64 );  //Low speed

    _card_disable(); 
}

/*******************************************************************************
* Function Name  : MSD_Init
* Description    : SD Card initializtion
* Input          : None
* Output         : None
* Return         : None
* Attention         : None
*******************************************************************************/
int MSD_Init(void)
{
    uint8_t r1;
    uint8_t buff[6] = {0};
    uint16_t retry;

    MSD_SPI_Configuration();

    /* Check , if no card insert */
    if( _card_insert() )
    {
#ifdef PRINT_INFO 
        PRINT("There is no card detected! \r\n");
#endif
      /* FATFS error flag */
      return -1;
    }

    /* Power on and delay some times */
    for(retry=0; retry<0x100; retry++)
    {
        _card_power_on();
    }

    /* Satrt send 74 clocks at least */
    for(retry=0; retry<10; retry++)
    {
        _spi_read_write(DUMMY_BYTE);
    }

    /* Start send CMD0 till return 0x01 means in IDLE state */
    for(retry=0; retry<0xFFF; retry++)
    {
        r1 = _send_command(CMD0, 0, 0x95);
        if(r1 == 0x01)
        {
            retry = 0;
            break;
        }
    }
    /* Timeout return */
    if(retry == 0xFFF)
    {
#ifdef PRINT_INFO 
        PRINT("Reset card into IDLE state failed!\r\n");
#endif
        return 1;
    }

    /* Get the card type, version */
    r1 = _send_command_hold(CMD8, 0x1AA, 0x87);
    /* r1=0x05 -> V1.0 */
    if(r1 == 0x05)
    {
      cardInfo.CardType = CARDTYPE_SDV1;

      /* End of CMD8, chip disable and dummy byte */
      _card_disable();
      _spi_read_write(DUMMY_BYTE);

      /* SD1.0/MMC start initialize */
      /* Send CMD55+ACMD41, No-response is a MMC card, otherwise is a SD1.0 card */
      for(retry=0; retry<0xFFF; retry++)
      {
         r1 = _send_command(CMD55, 0, 0);            /* should be return 0x01 */
         if(r1 != 0x01)
         {
#ifdef PRINT_INFO 
            PRINT("Send CMD55 should return 0x01, response=0x%02x\r\n", r1);
#endif
            return r1;
         }

         r1 = _send_command(ACMD41, 0, 0);            /* should be return 0x00 */
         if(r1 == 0x00)
         {
            retry = 0;
            break;
         }
      }

      /* MMC card initialize start */
      if(retry == 0xFFF)
      {
         for(retry=0; retry<0xFFF; retry++)
         {
             r1 = _send_command(CMD1, 0, 0);        /* should be return 0x00 */
             if(r1 == 0x00)
             {
                retry = 0;
                break;
             }
         }

         /* Timeout return */
         if(retry == 0xFFF)
         {
#ifdef PRINT_INFO
            PRINT("Send CMD1 should return 0x00, response=0x%02x\r\n", r1);
#endif
            return 2;
         }

        cardInfo.CardType = CARDTYPE_MMC;
#ifdef PRINT_INFO 
        PRINT("Card Type 111                 : MMC\r\n");
#endif
      }
        /* SD1.0 card detected, print information */
#ifdef PRINT_INFO
      else
      {
         PRINT("Card Type 111                : SD V1\r\n");
      }
#endif 

      /* Set spi speed high */
      SPI0_CLKCfg(4);

      /* CRC disable */
      r1 = _send_command(CMD59, 0, 0x01);
      if(r1 != 0x00)
      {
#ifdef PRINT_INFO 
          PRINT("Send CMD59 should return 0x00, response=0x%02x\r\n", r1);
#endif
          return r1;        /* response error, return r1 */
      }

      /* Set the block size */
      r1 = _send_command(CMD16, MSD_BLOCKSIZE, 0xFF);
      if(r1 != 0x00)
      {
#ifdef PRINT_INFO
          PRINT("Send CMD16 should return 0x00, response=0x%02x\r\n", r1);
#endif
          return r1;        /* response error, return r1 */
      }
   }

   /* r1=0x01 -> V2.x, read OCR register, check version */
   else if(r1 == 0x01)
   {
     /* 4Bytes returned after CMD8 sent    */
     buff[0] = _spi_read_write(DUMMY_BYTE);                /* should be 0x00 */
     buff[1] = _spi_read_write(DUMMY_BYTE);                /* should be 0x00 */
     buff[2] = _spi_read_write(DUMMY_BYTE);                /* should be 0x01 */
     buff[3] = _spi_read_write(DUMMY_BYTE);                /* should be 0xAA */

     /* End of CMD8, chip disable and dummy byte */
     _card_disable();
     _spi_read_write(DUMMY_BYTE);

     /* Check voltage range be 2.7-3.6V    */
     if(buff[2]==0x01 && buff[3]==0xAA)
     {
        for(retry=0; retry<0xFFF; retry++)
        {
            r1 = _send_command(CMD55, 0, 0);            /* should be return 0x01 */
            if(r1!=0x01)
            {
#ifdef PRINT_INFO 
                PRINT("Send CMD55 should return 0x01, response=0x%02x\r\n", r1);
#endif
                return r1;
            }

            r1 = _send_command(ACMD41, 0x40000000, 0);    /* should be return 0x00 */
            if(r1 == 0x00)
            {
                retry = 0;
                break;
            }
        }

        /* Timeout return */
        if(retry == 0xFFF)
        {
#ifdef PRINT_INFO
            PRINT("Send ACMD41 should return 0x00, response=0x%02x\r\n", r1);
#endif
            return 3;
        }

        /* Read OCR by CMD58 */
        r1 = _send_command_hold(CMD58, 0, 0);
        if(r1!=0x00)
        {
#ifdef PRINT_INFO
            PRINT("Send CMD58 should return 0x00, response=0x%02x\r\n", r1);
#endif
            return r1;        /* response error, return r1 */
        }

        buff[0] = _spi_read_write(DUMMY_BYTE);
        buff[1] = _spi_read_write(DUMMY_BYTE);
        buff[2] = _spi_read_write(DUMMY_BYTE);
        buff[3] = _spi_read_write(DUMMY_BYTE);

        /* End of CMD58, chip disable and dummy byte */
        _card_disable();
        _spi_read_write(DUMMY_BYTE);

        /* OCR -> CCS(bit30)  1: SDV2HC     0: SDV2 */
        if(buff[0] & 0x40)
        {
           cardInfo.CardType = CARDTYPE_SDV2HC;
#ifdef PRINT_INFO 
           PRINT("Card Type 222                 : SD V2HC\r\n");
#endif
        }
        else
        {
           cardInfo.CardType = CARDTYPE_SDV2;
#ifdef PRINT_INFO
           PRINT("Card Type 222                 : SD V2\r\n");
#endif
        }

        /* Set spi speed high */
        SPI0_CLKCfg(4);
     }
   }
   return 0;
}

/*******************************************************************************
* Function Name  : MSD_GetCardInfo
* Description    : Get SD Card Information
* Input          : None
* Output         : None
* Return         : 0：NO_ERR; TRUE: Error
* Attention         : None
*******************************************************************************/
int MSD_GetCardInfo(PMSD_CARDINFO cardInfo)
{
    uint8_t r1,i;
    uint8_t CSD_Tab[16];
    uint8_t CID_Tab[16];

    /* Send CMD9, Read CSD */
    r1 = _send_command_hold(CMD9, 0, 0xFF);
    if(r1 != 0x00)
    {
  #ifdef PRINT_INFO
      PRINT("CMD9 resp error %X.\r\n",r1);
  #endif
      return r1;
    }

    if(_read_buffer(CSD_Tab, 16, RELEASE))
    {
  #ifdef PRINT_INFO
      PRINT("CMD9 CSD_Tab error.\r\n");
  #endif
      return 1;
    }

    /* Send CMD10, Read CID */
    r1 = _send_command_hold(CMD10, 0, 0xFF);
    if(r1 != 0x00)
    {
  #ifdef PRINT_INFO
      PRINT("CMD10 resp error %X.\r\n",r1);
  #endif
      return r1;
    }

    if(_read_buffer(CID_Tab, 16, RELEASE))
    {
  #ifdef PRINT_INFO
      PRINT("CMD10 CID_Tab error.\r\n");
  #endif
      return 2;
    }

  #ifdef PRINT_INFO
    PRINT("CSD_Tab:");
    for(i=0;i<16;i++)
    {
        PRINT("%02X ",CSD_Tab[i]);
    }
    PRINT("\r\n");
  #endif

    /* Byte 0 */
    cardInfo->CSD.CSDStruct = (CSD_Tab[0] & 0xC0) >> 6;
    cardInfo->CSD.SysSpecVersion = (CSD_Tab[0] & 0x3C) >> 2;
    cardInfo->CSD.Reserved1 = CSD_Tab[0] & 0x03;
    /* Byte 1 */
    cardInfo->CSD.TAAC = CSD_Tab[1] ;
    /* Byte 2 */
    cardInfo->CSD.NSAC = CSD_Tab[2];
    /* Byte 3 */
    cardInfo->CSD.MaxBusClkFrec = CSD_Tab[3];
    /* Byte 4 */
    cardInfo->CSD.CardComdClasses = CSD_Tab[4] << 4;
    /* Byte 5 */
    cardInfo->CSD.CardComdClasses |= (CSD_Tab[5] & 0xF0) >> 4;
    cardInfo->CSD.RdBlockLen = CSD_Tab[5] & 0x0F;
    /* Byte 6 */
    cardInfo->CSD.PartBlockRead = (CSD_Tab[6] & 0x80) >> 7;
    cardInfo->CSD.WrBlockMisalign = (CSD_Tab[6] & 0x40) >> 6;
    cardInfo->CSD.RdBlockMisalign = (CSD_Tab[6] & 0x20) >> 5;
    cardInfo->CSD.DSRImpl = (CSD_Tab[6] & 0x10) >> 4;
    cardInfo->CSD.Reserved2 = 0; /* Reserved */
    cardInfo->CSD.DeviceSize = (CSD_Tab[6] & 0x03) << 10;
    /* Byte 7 */
    cardInfo->CSD.DeviceSize |= (CSD_Tab[7]) << 2;
    /* Byte 8 */
    cardInfo->CSD.DeviceSize |= (CSD_Tab[8] & 0xC0) >> 6;
    cardInfo->CSD.MaxRdCurrentVDDMin = (CSD_Tab[8] & 0x38) >> 3;
    cardInfo->CSD.MaxRdCurrentVDDMax = (CSD_Tab[8] & 0x07);
    /* Byte 9 */
    cardInfo->CSD.MaxWrCurrentVDDMin = (CSD_Tab[9] & 0xE0) >> 5;
    cardInfo->CSD.MaxWrCurrentVDDMax = (CSD_Tab[9] & 0x1C) >> 2;
    cardInfo->CSD.DeviceSizeMul = (CSD_Tab[9] & 0x03) << 1;
    /* Byte 10 */
    cardInfo->CSD.DeviceSizeMul |= (CSD_Tab[10] & 0x80) >> 7;
    cardInfo->CSD.EraseGrSize = (CSD_Tab[10] & 0x7C) >> 2;
    cardInfo->CSD.EraseGrMul = (CSD_Tab[10] & 0x03) << 3;
    /* Byte 11 */
    cardInfo->CSD.EraseGrMul |= (CSD_Tab[11] & 0xE0) >> 5;
    cardInfo->CSD.WrProtectGrSize = (CSD_Tab[11] & 0x1F);
    /* Byte 12 */
    cardInfo->CSD.WrProtectGrEnable = (CSD_Tab[12] & 0x80) >> 7;
    cardInfo->CSD.ManDeflECC = (CSD_Tab[12] & 0x60) >> 5;
    cardInfo->CSD.WrSpeedFact = (CSD_Tab[12] & 0x1C) >> 2;
    cardInfo->CSD.MaxWrBlockLen = (CSD_Tab[12] & 0x03) << 2;
    /* Byte 13 */
    cardInfo->CSD.MaxWrBlockLen |= (CSD_Tab[13] & 0xc0) >> 6;
    cardInfo->CSD.WriteBlockPaPartial = (CSD_Tab[13] & 0x20) >> 5;
    cardInfo->CSD.Reserved3 = 0;
    cardInfo->CSD.ContentProtectAppli = (CSD_Tab[13] & 0x01);
    /* Byte 14 */
    cardInfo->CSD.FileFormatGrouop = (CSD_Tab[14] & 0x80) >> 7;
    cardInfo->CSD.CopyFlag = (CSD_Tab[14] & 0x40) >> 6;
    cardInfo->CSD.PermWrProtect = (CSD_Tab[14] & 0x20) >> 5;
    cardInfo->CSD.TempWrProtect = (CSD_Tab[14] & 0x10) >> 4;
    cardInfo->CSD.FileFormat = (CSD_Tab[14] & 0x0C) >> 2;
    cardInfo->CSD.ECC = (CSD_Tab[14] & 0x03);
    /* Byte 15 */
    cardInfo->CSD.CSD_CRC = (CSD_Tab[15] & 0xFE) >> 1;
    cardInfo->CSD.Reserved4 = 1;
  
    if(cardInfo->CardType == CARDTYPE_SDV2HC)
    {
       /* Byte 7 */
       cardInfo->CSD.DeviceSize = (u16)(CSD_Tab[8]) *256;
       /* Byte 8 */
       cardInfo->CSD.DeviceSize += CSD_Tab[9] ;
    }
  
    cardInfo->Capacity = cardInfo->CSD.DeviceSize * MSD_BLOCKSIZE * 1024;
    cardInfo->BlockSize = MSD_BLOCKSIZE;

  #ifdef PRINT_INFO
    PRINT("Capacity=0x%X, BlockSize=0x%X.\r\n",cardInfo->Capacity,cardInfo->BlockSize);
    PRINT("CID_Tab:");
    for(i=0;i<16;i++)
    {
        PRINT("%02X ",CID_Tab[i]);
    }
    PRINT("\r\n");
  #endif

    /* Byte 0 */
    cardInfo->CID.ManufacturerID = CID_Tab[0];
    /* Byte 1 */
    cardInfo->CID.OEM_AppliID = CID_Tab[1] << 8;
    /* Byte 2 */
    cardInfo->CID.OEM_AppliID |= CID_Tab[2];
    /* Byte 3 */
    cardInfo->CID.ProdName1 = CID_Tab[3] << 24;
    /* Byte 4 */
    cardInfo->CID.ProdName1 |= CID_Tab[4] << 16;
    /* Byte 5 */
    cardInfo->CID.ProdName1 |= CID_Tab[5] << 8;
    /* Byte 6 */
    cardInfo->CID.ProdName1 |= CID_Tab[6];
    /* Byte 7 */
    cardInfo->CID.ProdName2 = CID_Tab[7];
    /* Byte 8 */
    cardInfo->CID.ProdRev = CID_Tab[8];
    /* Byte 9 */
    cardInfo->CID.ProdSN = CID_Tab[9] << 24;
    /* Byte 10 */
    cardInfo->CID.ProdSN |= CID_Tab[10] << 16;
    /* Byte 11 */
    cardInfo->CID.ProdSN |= CID_Tab[11] << 8;
    /* Byte 12 */
    cardInfo->CID.ProdSN |= CID_Tab[12];
    /* Byte 13 */
    cardInfo->CID.Reserved1 |= (CID_Tab[13] & 0xF0) >> 4;
    /* Byte 14 */
    cardInfo->CID.ManufactDate = (CID_Tab[13] & 0x0F) << 8;
    /* Byte 15 */
    cardInfo->CID.ManufactDate |= CID_Tab[14];
    /* Byte 16 */
    cardInfo->CID.CID_CRC = (CID_Tab[15] & 0xFE) >> 1;
    cardInfo->CID.Reserved2 = 1;
  
    return 0;
}

/*******************************************************************************
* Function Name  : _read_buffer
* Description    : None
* Input          : - *buff:
*                   - len:
*                   - release:
* Output         : None
* Return         : 0：NO_ERR; TRUE: Error
* Attention         : None
*******************************************************************************/
int _read_buffer(uint8_t *buff, uint16_t len, uint8_t release)
{
    uint8_t r1;
    uint16_t retry;

    /* Card enable, Prepare to read    */
    _card_enable();

    /* Wait start-token 0xFE */
    for(retry=0; retry<2000; retry++)
    {
       r1 = _spi_read_write(DUMMY_BYTE);
       if(r1 == 0xFE)
       {
           retry = 0;
           break;
       }
    }

    /* Timeout return    */
    if(retry == 2000)
    {
  #ifdef PRINT_INFO
        PRINT("_read_buffer Timeout return.\r\n");
  #endif
       _card_disable();
       return 1;
    }

    /* Start reading */
    for(retry=0; retry<len; retry++)
    {
       *(buff+retry) = _spi_read_write(DUMMY_BYTE);
    }

    /* 2bytes dummy CRC */
    _spi_read_write(DUMMY_BYTE);
    _spi_read_write(DUMMY_BYTE);

    /* chip disable and dummy byte */
    if(release)
    {
       _card_disable();
       _spi_read_write(DUMMY_BYTE);
    }

    return 0;
}

/*******************************************************************************
* Function Name  : MSD_ReadSingleBlock
* Description    : None
* Input          : - sector:
*                   - buffer:
* Output         : None
* Return         : None
* Attention         : None
*******************************************************************************/
int MSD_ReadSingleBlock(uint32_t sector, uint8_t *buffer)
{
    uint8_t r1;

    /* if ver = SD2.0 HC, sector need <<9 */
    if(cardInfo.CardType != CARDTYPE_SDV2HC)
    {
       sector = sector<<9;
    }

    /* Send CMD17 : Read single block command */
    r1 = _send_command_hold(CMD17, sector, 0);
    if(r1 != 0x00)
    {
       return 1;
    }

    /* Start read and return the result */
    r1 = _read_buffer(buffer, MSD_BLOCKSIZE, RELEASE);   //  HOLD
  
    /* Send stop data transmit command - CMD12 */
    _send_command(CMD12, 0, 0);

    return r1;
}

/*******************************************************************************
* Function Name  : MSD_ReadMultiBlock
* Description    : None
* Input          : - sector:
*                   - buffer:
*                  - NbrOfSector:
* Output         : None
* Return         : None
* Attention         : None
*******************************************************************************/
int MSD_ReadMultiBlock(uint32_t sector, uint8_t *buffer, uint32_t NbrOfSector)
{
    uint8_t r1;
    uint32_t i;

    /* if ver = SD2.0 HC, sector need <<9 */
    if(cardInfo.CardType != CARDTYPE_SDV2HC)
    {
       sector = sector<<9;
    }

    /* Send CMD18 : Read multi block command */
    r1 = _send_command_hold(CMD18, sector, 0);
    if(r1 != 0x00)
    {
       return 1;
    }

    /* Start read    */
    for(i=0; i<NbrOfSector; i++)
    {
       if(_read_buffer(buffer+i*MSD_BLOCKSIZE, MSD_BLOCKSIZE, HOLD))
       {
           /* Send stop data transmit command - CMD12    */
           _send_command(CMD12, 0, 0);
           /* chip disable and dummy byte */
           _card_disable();
           return 2;
       }
    }

    /* Send stop data transmit command - CMD12 */
    _send_command(CMD12, 0, 0);

    /* chip disable and dummy byte */
    _card_disable();
    _spi_read_write(DUMMY_BYTE);

    return 0;
}

/*******************************************************************************
* Function Name  : MSD_WriteSingleBlock
* Description    : None
* Input          : - sector:
*                   - buffer:
* Output         : None
* Return         : None
* Attention         : None
*******************************************************************************/
int MSD_WriteSingleBlock(uint32_t sector, uint8_t *buffer)
{
    uint8_t r1;
    uint16_t i;
    uint32_t retry;

    /* if ver = SD2.0 HC, sector need <<9 */
    if(cardInfo.CardType != CARDTYPE_SDV2HC)
    {
       sector = sector<<9;
    }

    /* Send CMD24 : Write single block command */
    r1 = _send_command(CMD24, sector, 0);

    if(r1 != 0x00)
    {
       return 1;
    }

    /* Card enable, Prepare to write */
    _card_enable();
    _spi_read_write(DUMMY_BYTE);
    _spi_read_write(DUMMY_BYTE);
    _spi_read_write(DUMMY_BYTE);
    /* Start data write token: 0xFE */
    _spi_read_write(0xFE);

    /* Start single block write the data buffer */
    for(i=0; i<MSD_BLOCKSIZE; i++)
    {
      _spi_read_write(*buffer++);
    }

    /* 2Bytes dummy CRC */
    _spi_read_write(DUMMY_BYTE);
    _spi_read_write(DUMMY_BYTE);

    /* MSD card accept the data */
    r1 = _spi_read_write(DUMMY_BYTE);
    if((r1&0x1F) != 0x05)
    {
      _card_disable();
      return 2;
    }

    /* Wait all the data programm finished */
    retry = 0;
    while(_spi_read_write(DUMMY_BYTE) == 0x00)
    {
       /* Timeout return */
       if(retry++ == 0x40000)
       {
          _card_disable();
          return 3;
       }
    }

    /* chip disable and dummy byte */
    _card_disable();
    _spi_read_write(DUMMY_BYTE);

    return 0;
}

/*******************************************************************************
* Function Name  : MSD_WriteMultiBlock
* Description    : None
* Input          : - sector:
*                   - buffer:
*                  - NbrOfSector:
* Output         : None
* Return         : None
* Attention         : None
*******************************************************************************/
int MSD_WriteMultiBlock(uint32_t sector, uint8_t *buffer, uint32_t NbrOfSector)
{
    uint8_t r1;
    uint16_t i;
    uint32_t n;
    uint32_t retry;

    /* if ver = SD2.0 HC, sector need <<9 */
    if(cardInfo.CardType != CARDTYPE_SDV2HC)
    {
        sector = sector<<9;
    }

    /* Send command ACMD23 berfore multi write if is not a MMC card */
    if(cardInfo.CardType != CARDTYPE_MMC)
    {
        _send_command(ACMD23, NbrOfSector, 0x00);
    }

    /* Send CMD25 : Write nulti block command    */
    r1 = _send_command(CMD25, sector, 0);

    if(r1 != 0x00)
    {
        return 1;
    }

    /* Card enable, Prepare to write */
    _card_enable();
    _spi_read_write(DUMMY_BYTE);
    _spi_read_write(DUMMY_BYTE);
    _spi_read_write(DUMMY_BYTE);

    for(n=0; n<NbrOfSector; n++)
    {
       /* Start multi block write token: 0xFC */
       _spi_read_write(0xFC);

       for(i=0; i<MSD_BLOCKSIZE; i++)
       {
          _spi_read_write(*buffer++);
       }

       /* 2Bytes dummy CRC */
       _spi_read_write(DUMMY_BYTE);
       _spi_read_write(DUMMY_BYTE);

       /* MSD card accept the data */
       r1 = _spi_read_write(DUMMY_BYTE);
       if((r1&0x1F) != 0x05)
       {
          _card_disable();
          return 2;
       }

       /* Wait all the data programm finished    */
       retry = 0;
       while(_spi_read_write(DUMMY_BYTE) != 0xFF)
       {
          /* Timeout return */
          if(retry++ == 0x40000)
          {
             _card_disable();
             return 3;
          }
       }
    }

    /* Send end of transmit token: 0xFD */
    r1 = _spi_read_write(0xFD);
    if(r1 == 0x00)
    {
       return 4;
    }

    /* Wait all the data programm finished */
    retry = 0;
    while(_spi_read_write(DUMMY_BYTE) != 0xFF)
    {
       /* Timeout return */
       if(retry++ == 0x40000)
       {
           _card_disable();
           return 5;
       }
    }

    /* chip disable and dummy byte */
    _card_disable();
    _spi_read_write(DUMMY_BYTE);

    return 0;
}

/*******************************************************************************
* Function Name  : _send_command
* Description    : None
* Input          : - cmd:
*                   - arg:
*                  - crc:
* Output         : None
* Return         : R1 value, response from card
* Attention         : None
*******************************************************************************/
int _send_command(uint8_t cmd, uint32_t arg, uint8_t crc)
{
    uint8_t r1;
    uint8_t retry;

    /* Dummy byte and chip enable */
    _spi_read_write(DUMMY_BYTE);
    _card_enable();

    /* Command, argument and crc */
    _spi_read_write(cmd | 0x40);
    _spi_read_write(arg >> 24);
    _spi_read_write(arg >> 16);
    _spi_read_write(arg >> 8);
    _spi_read_write(arg);
    _spi_read_write(crc);

    /* Wait response, quit till timeout */
    for(retry=0; retry<200; retry++)
    {
       r1 = _spi_read_write(DUMMY_BYTE);
       if(r1 != 0xFF)
       {
           break;
       }
    }

    /* Chip disable and dummy byte */
    _card_disable();
    _spi_read_write(DUMMY_BYTE);
  
    return r1;
}

/*******************************************************************************
* Function Name  : _send_command_hold
* Description    : None
* Input          : - cmd:
*                   - arg:
*                  - crc:
* Output         : None
* Return         : R1 value, response from card
* Attention         : None
*******************************************************************************/
int _send_command_hold(uint8_t cmd, uint32_t arg, uint8_t crc)
{
    uint8_t r1;
    uint8_t retry;

    /* Dummy byte and chip enable */
    _spi_read_write(DUMMY_BYTE);
    _card_enable();

    /* Command, argument and crc */
    _spi_read_write(cmd | 0x40);
    _spi_read_write(arg >> 24);
    _spi_read_write(arg >> 16);
    _spi_read_write(arg >> 8);
    _spi_read_write(arg);
    _spi_read_write(crc);

    /* Wait response, quit till timeout */
    for(retry=0; retry<200; retry++)
    {
       r1 = _spi_read_write(DUMMY_BYTE);
       if(r1 != 0xFF)
       {
           break;
       }
    }

    return r1;
}



/*******************************************************************************
* Function Name  : scan_files
* Description    : 搜索文件目录下所有文件
* Input          : - path: 根目录
* Output         : None
* Return         : FRESULT
* Attention    : 不支持长文件名
*******************************************************************************/
FRESULT scan_files(char* path)
{
    FILINFO fno;
    DIR dir;
    int i;
    char *fn;

#if _USE_LFN
    static char lfn[_MAX_LFN * 2 + 1];
    fno.lfname = lfn;
    fno.lfsize = sizeof(lfn);
#endif

    res = f_opendir(&dir, path);
    if (res == FR_OK)
    {
        i = strlen(path);
        for (;;)
        {
            res = f_readdir(&dir, &fno);
            if (res != FR_OK || fno.fname[0] == 0) break;
            if (fno.fname[0] == '.') continue;
#if _USE_LFN
            fn = *fno.lfname ? fno.lfname : fno.fname;
#else
            fn = fno.fname;
#endif
            if (fno.fattrib & AM_DIR)
            {
                sprintf(&path[i], "/%s", fn);
                PRINT("DIR:%s \r\n",path);
                res = scan_files(path);
                if (res != FR_OK) break;
                path[i] = 0;
            } else {
                PRINT("%s/%s \r\n", path, fn);
            }
        }

    }else   PRINT("f_opendir %s error!\r\n",path);

    PRINT("scan_files over! RES=%d.\r\n",res);
    return res;
}
/*******************************************************************************
* Function Name  : SD_TotalSize
* Description    : 文件空间占用情况
* Input          : None
* Output         : None
* Return         : 返回1成功 返回0失败
* Attention    : None
*******************************************************************************/
int SD_TotalSize(void)
{
    FATFS *fs;
    DWORD fre_clust;

    res = f_getfree("0:", &fre_clust, &fs);  /* 必须是根目录，选择磁盘0 */
    if ( res==FR_OK )
    {
    /* Print free space in unit of MB (assuming 512 bytes/sector) */
      PRINT("\r\n%d MB total drive space.\r\n"
           "%d MB available.\r\n",
           ( (fs->n_fatent - 2) * fs->csize ) / 2 /1024 , (fre_clust * fs->csize) / 2 /1024 );

    return ENABLE;
  }else   return DISABLE;
}
/*******************************************************************************/
void SDCARD_FF_Init(){
    UINT32       i,j;
    uint8_t      *p;

    if(MSD_Init())
    {
        PRINT("error!\r\n");
    }else
    {
        MSD_GetCardInfo(&cardInfo);
        PRINT("OK\r\n");
#ifdef DEBUG_INFO
        if(MSD_ReadSingleBlock(0,buf) == 0)
        {
            PRINT("TF Card Sector Data:\r\n");
            p = buf;
            for(i=0; i!=512/16; i++)
            {
                for(j=0; j!=16; j++)
                {
                    PRINT("%02X ",*p++);
                }
                PRINT("\r\n");
            }
        }

//        if(MSD_WriteSingleBlock(1,buf) == 0)
//        {
//            PRINT("TF Card Sector Data:\r\n");
//            p = buf;
//            for(i=0; i!=512/16; i++)
//            {
//                for(j=0; j!=16; j++)
//                {
//                    PRINT("%02X ",*p++);
//                }
//                PRINT("\r\n");
//            }
//        }
//        if(MSD_ReadSingleBlock(1,buf) == 0)
//        {
//            PRINT("TF Card Sector Data:\r\n");
//            p = buf;
//            for(i=0; i!=512/16; i++)
//            {
//                for(j=0; j!=16; j++)
//                {
//                    PRINT("%02X ",*p++);
//                }
//                PRINT("\r\n");
//            }
//        }

//        if(MSD_ReadMultiBlock(0,buf,2) == 0)
//        {
//            PRINT("TF Card Sector Data:\r\n");
//            p = buf;
//            for(i=0; i!=1024/16; i++)
//            {
//                for(j=0; j!=16; j++)
//                {
//                    PRINT("%02X ",*p++);
//                }
//                PRINT("\r\n");
//            }
//        }

        res = f_open( &fsrc , "0:/Demo.TXT" , FA_CREATE_NEW | FA_WRITE);
        if ( res == FR_OK )
        {
            /* Write buffer to file */
            res = f_write(&fsrc, textFileBuffer, sizeof(textFileBuffer), &br);

            PRINT("Demo.TXT successfully created        \r\n");

            /*close file */
            f_close(&fsrc);
        }
        else if ( res == FR_EXIST )
        {
            PRINT("Demo.TXT created in the disk      \r\n");
        }else
        {
            PRINT("Demo.TXT created in the disk error.\r\n");
        }
#endif
        f_mount(&fs,"0:",1);
        scan_files(path);
        SD_TotalSize();
    }
}
